-- Convenience Lua functions that can be used within rpm macros

-- Reads an rpm variable. Unlike a basic rpm.expand("{?foo}"), returns nil if
-- the variable is unset, which is convenient in lua tests and enables
-- differentiating unset variables from variables set to ""
local function read(rpmvar)
  if not rpmvar or
    (rpm.expand("%{" .. rpmvar .. "}") == "%{" .. rpmvar .. "}") then
    return nil
  else
    return rpm.expand("%{?" .. rpmvar .. "}")
  end
end

-- Builds a list of un-commented and non-empty lines from a multiline rpm
-- variable. This allows using multiline expands for private %sourcelist-like
-- containers.
local function readlines(rpmvar)
  local lines = read(rpmvar)
  local L = {}
  if lines then
    for line in string.gmatch(lines,'[^\r\n]+') do
      if string.match(line, '^[^#]') then
        table.insert(L, line)
      end
    end
  end
  return L
end

-- Returns true if the macro that called this function had flag set
--   – for example, hasflag("z") would give the following results:
--     %foo -z bar → true
--     %foo -z     → true
--     %foo        → false
local function hasflag(flag)
  return (rpm.expand("%{-" .. flag .. "}") ~= "")
end

-- Returns the argument passed to flag in the macro that called this function
--  – for example, readflag("z") would give the following results:
--      %foo -z bar → bar
--      %foo        → nil
--      %foo -z ""  → empty string
--      %foo -z ''  → empty string
--  – typically used to feed index to floop
local function readflag(flag)
  if not hasflag(flag) then
    return nil
  else
    local a = rpm.expand("%{-" .. flag .. "*}")
    -- Handle "" and '' as empty strings
    if (a == '""') or (a == "''") then
      a = ''
    end
    return a
  end
end

-- %wordwrap core
local function wordwrap(text)
  text = rpm.expand(text .. "\n")
  text = text:gsub("\t",               "  ")
  text = text:gsub("\r",               "\n")
  text = text:gsub(" +\n",             "\n")
  text = text:gsub("\n+\n",            "\n\n")
  text = text:gsub("^\n",              "")
  text = text:gsub("\n( *)[-*—][  ]+", "\n%1– ")
  output = ""
  for line in text:gmatch("[^\n]*\n") do
    local pos = 0
    local advance = ""
    for word in line:gmatch("%s*[^%s]*\n?") do
      local wl, bad = utf8.len(word)
      if not wl then
        -- Can not use warning in wordwrap since warning uses wordwrap
        rpm.expand([[
%{warn:Invalid UTF-8 sequence detected in:}
%{warn:]] .. word .. [[}
%{warn:It may produce unexpected results.}
]])
        wl = bad
      end
      if (pos == 0) then
        advance, n = word:gsub("^(%s*– ).*", "%1")
        if (n == 0) then
          advance  = word:gsub("^(%s*).*",   "%1")
        end
        advance = advance:gsub("– ", "  ")
        pos = pos + wl
      elseif  (pos + wl  < 81) or
             ((pos + wl == 81) and word:match("\n$")) then
        pos = pos + wl
      else
        word = advance .. word:gsub("^%s*", "")
        output = output .. "\n"
        pos = utf8.len(word)
      end
      output = output .. word
      if pos > 80 then
        pos = 0
        if not word:match("\n$") then
          output = output .. "\n"
        end
      end
    end
  end
  output = output:gsub("^\n*", "")
  output = output:gsub("\n*$", "\n")
  -- Escape %’s – we want macros that were preserved from wordwrap expansion to
  -- be preserved at the next stage too
  output = output:gsub("%%", "%%%%")
  return output
end

-- Reformats text and outputs it using a built-in rpm verb such as echo, warn
-- or error
local function message(verb, text)
  rpm.expand("%{" .. verb .. ":" .. wordwrap(text):gsub("\n*$", "") .. "}")
end

-- Writes some text to stdout
local function echo(text)
  message("echo", text)
end

-- Writes the list of z-suffixed rpmvars to the console, if set
local function echovars(rpmvars, z)
  for _, rpmvar in ipairs(rpmvars) do
    local suffixed = rpmvar .. z
    local header = string.sub("  " .. suffixed .. ":                                               ",1,24)
    if read(suffixed) then
      echo(header .. "%{" .. suffixed .. "}")
    end
  end
end

-- Writes some text as a warning
local function warning(text)
  message("warn", text)
end

-- Writes some text as an error
local function err(text)
  message("error", text)
end

-- Sets a spec variable; echoes the result if verbose
local function set(rpmvar, value, verbose)
  if (value == nil) then
    value = "%{nil}"
  end
  if verbose then
    if (value == "%{nil}") then
      warning("Unsetting " .. rpmvar)
    else
      warning("Setting "   .. rpmvar .. " = " .. value)
    end
  end
  rpm.define(rpmvar .. " " .. value)
end

-- Unsets a spec variable if it is defined; echoes the result if verbose
local function unset(rpmvar, verbose)
  if read(rpmvar) then
    set(rpmvar, nil, verbose)
  end
end

-- Sets a spec variable, if not already set; echoes the result if verbose
local function safeset(rpmvar, value, verbose)
  if not read(rpmvar) then
    set(rpmvar, value, verbose)
  end
end

-- Aliases a list of rpm variables to the same variables suffixed with 0 (and
-- vice versa); echoes the result if verbose
local function zalias(rpmvars, verbose)
  for _, sfx in ipairs({{"","0"},{"0",""}}) do
    for _, rpmvar in ipairs(rpmvars) do
      local toalias = rpmvar .. sfx[1]
      if read(toalias) then
        safeset(rpmvar .. sfx[2], "%{" .. toalias .. "}", verbose)
      end
    end
  end
end

-- Takes a list of rpm variable roots and a suffix and alias __current_<rpmvar>
-- to <rpmvar><z> if it resolves to something not empty
local function set_current(rpmvars, z, verbose)
  for _, rpmvar in ipairs(rpmvars) do
    local suffixed = rpmvar .. z
    if read(suffixed) then
      set(  "__current_" .. rpmvar, "%{" .. suffixed .. "}", verbose)
    else
      unset("__current_" .. rpmvar,                          verbose)
    end
  end
end

-- Sets the usual verbosity variables in a generic way
local function set_verbose(verbose)
  if verbose then
    set(  "__current_verbose",       "-v",     verbose)
    unset("__current_quiet",                   verbose)
    unset("__current_shell_quiet",             verbose)
    unset("__current_shell_verbose",           verbose)
  else
    unset("__current_verbose",                 verbose)
    set(  "__current_quiet",         "-q",     verbose)
    set(  "__current_shell_quiet",   "set +x", verbose)
    set(  "__current_shell_verbose", "set -x", verbose)
  end
end

-- Merges two lists, removing duplicates and ordering the result
-- The result is a list containing an ordered set of values
local function mergelists(list_of_lists)
  local S = {}
  local L = {}
  for   _, l in ipairs(list_of_lists) do
    for _, v in ipairs(l) do
      S[v] = true
    end
  end
  for k, _ in  pairs(S) do
    table.insert(L, k)
  end
  table.sort(L)
  return L
end

-- Returns an array, indexed by suffix, containing the non-empy values of
-- <rpmvar><suffix>, with suffix an integer string or the empty string
local function suffixed(rpmvar)
  local S = {}
  zalias({rpmvar})
  for z=0,9999 do
    local value = read(rpmvar .. z)
    if value then
      S[tostring(z)] = value
    end
  end
  -- rpm convention is to alias no suffix to zero suffix
  -- only add no suffix if zero suffix is different
  local value = read(rpmvar)
  if value and (value ~= S["0"]) then
    S[""] = value
  end
  return S
end

-- Returns the list of suffixes, including the empty string, for which
-- <rpmvar><suffix> is set to a non empty value
local function suffixes(rpmvar)
  local S = {}
  for z in pairs(suffixed(rpmvar)) do
    table.insert(S,z)
  end
  table.sort(S,
             function(a,b) return (tonumber(a) or 0) < (tonumber(b) or 0) end)
  return S
end

-- Returns the list of suffixes, including the empty string, for which
-- <rpmvar><suffix> is set to a non empty value, for any rpmvar in rpmvars
local function allsuffixes(rpmvars)
  local SL = {}
  local S  = {}
  for _,rpmvar in ipairs(rpmvars) do
    table.insert(SL, suffixes(rpmvar))
  end
  S = mergelists(SL)
  table.sort(S,
             function(a,b) return (tonumber(a) or 0) < (tonumber(b) or 0) end)
  return S
end

-- Returns the suffix for which <rpmvar><suffix> has a non-empty value that
-- matches best the beginning of the value string
local function bestsuffix(rpmvar, value)
  local best         = nil
  local currentmatch = ""
  for z, setvalue in pairs(suffixed(rpmvar)) do
  if ((not best) or (string.len(setvalue) < string.len(currentmatch))) and
     (string.find(setvalue, "^" .. value)) then
      currentmatch = setvalue
      best         = z
    end
  end
  return best
end

-- Finds an unclaimed suffix and set <rpmvar><suffix> to <value>
local function suffix_new(rpmvar, value, verbose)
  local s = 0
  local z = ""
  repeat
    z = tostring(s)
    s = s + 1
  until not read(rpmvar .. z)
  safeset(rpmvar .. z, value, verbose)
  return z
end

-- Returns a suffix for which <rpmvar><suffix> is set to <value>. If none
-- exists, find an available suffix and use it for a new entry
local function suffix_reuse(rpmvar, value, verbose)
  local z = bestsuffix(rpmvar, value)
  if z and (read(rpmvar .. z) == value) then
    return z
  else
    return suffix_new(rpmvar, value, verbose)
  end
end

-- Returns a suffix for which <rpmvar><suffix> is set to <value>, optionaly
-- reusing exisiting entries
local function suffix(rpmvar, value, reuse, verbose)
  if reuse then
    return suffix_reuse(rpmvar, value, verbose)
  else
    return suffix_new(rpmvar, value, verbose)
  end
end

-- Returns the patch suffix/number of a filename, if it has been declared
-- as a string, since it will be usually used to construct spec strings
local function patchsuffix(filename)
  local filepath = rpm.expand("%{_sourcedir}/" .. filename)
  for z, p in ipairs(patches) do
    if ( p == filepath ) then
      return tostring(z)
    end
  end
  return nil
end

-- Executes the f function in a loop
--   – the loop is controled by index (a string):
--     — if the index value given to floop is nil, loop over the whole
--       range. Use readflag() to process rpm flag arguments safely.
--     — otherwise execute f for the specified index value only.
--   – f is passed index then otherargs as arguments.
local function floop(f, index, range, otherargs)
  for _, i in pairs(range) do
    if (index == nil) or (i == index) then
      f(i, table.unpack(otherargs))
    end
  end
end

-- %writevars core
local function writevars(macrofile, rpmvars)
  for _, rpmvar in ipairs(rpmvars) do
    print("sed -i 's\029" .. string.upper("@@" .. rpmvar .. "@@") ..
                   "\029" .. rpm.expand(  "%{" .. rpmvar .. "}" ) ..
                   "\029g' " .. macrofile .. "\n")
  end
end

-- Because rpmbuild will fail if a subpackage is declared before the source
-- package itself, provide a source package declaration shell as fallback.
local function source_pkg(verbose)
  if verbose then
    echo([[
Creating a header for the SRPM from:
  – %%{source_name},
  – %%{source_summary},
  – %%{source_version},
  – %%{source_release},
  – %%{source_url},
  – %%{source_license} and
  – %%{source_description}.
If that is not the intended result, please declare the SRPM header and set
%%{source_name} in your spec file before calling a macro that creates other
package headers.
]])
  end
  print(rpm.expand([[
Name:           %{source_name}
Summary:        %{source_summary}
%{?source_version:Version:        %{source_version}}
%{?source_release:Release:        %{source_release}%{?dist}}
%{?source_license:License:        %{source_license}}
%{?source_url:URL:            %{source_url}}
%description
%wordwrap -v source_description
]]))
  set("__current_name", "%{source_name}", verbose)
end

-- %new_package core
local function new_package(source_name, pkg_name, name_suffix, first, verbose)
  -- Safety net when the wrapper is used in conjunction with traditional syntax
  if (not first) and (not source_name) then
    warning([[
Something already set a package name. However, %%{source_name} is not set.
Please set %%{source_name} to the SRPM name to ensure reliable processing.
]])
    if name_suffix then
      print(rpm.expand("%package        " .. name_suffix))
    else
      print(rpm.expand("%package     -n " .. pkg_name))
    end
    return
  end
  -- New processing
  if not (pkg_name or name_suffix or source_name) then
    err("You need to set %%{source_name} or provide explicit package naming!")
  end
  if name_suffix then
    print(rpm.expand("%package        "  .. name_suffix))
    set("__current_name", "%{source_name}-" .. name_suffix, verbose)
  else
    if not source_name then
      source_name = pkg_name
    end
    if (pkg_name == source_name) then
      safeset("source_name", source_name, verbose)
      print(rpm.expand("Name:           %{source_name}"))
    else
      if source_name and first then
        source_pkg(verbose)
      end
      print(rpm.expand("%package     -n " .. pkg_name))
    end
    set("__current_name", pkg_name, verbose)
  end
end

-- deprecated
local function setcurrent(rpmvars, suffix, verbose)
  warning("setcurrent is deprecated, use set_current instead")
  for _, rpmvar in ipairs(rpmvars) do
    local suffixed = rpmvar .. suffix
    if read(suffixed) then
      set(  "current" .. rpmvar, "%{" .. suffixed .. "}", verbose)
    else
      unset("current" .. rpmvar,                          verbose)
    end
  end
end

return {
  read           = read,
  readlines      = readlines,
  hasflag        = hasflag,
  readflag       = readflag,
  wordwrap       = wordwrap,
  echo           = echo,
  echovars       = echovars,
  warning        = warning,
  error          = err,
  set            = set,
  unset          = unset,
  safeset        = safeset,
  zalias         = zalias,
  set_current    = set_current,
  set_verbose    = set_verbose,
  mergelists     = mergelists,
  suffixed       = suffixed,
  suffixes       = suffixes,
  allsuffixes    = allsuffixes,
  bestsuffix     = bestsuffix,
  suffix         = suffix,
  patchsuffix    = patchsuffix,
  floop          = floop,
  writevars      = writevars,
  new_package    = new_package,
  -- deprecated aliases
  explicitset    = set,
  explicitunset  = unset,
  getsuffixed    = suffixed,
  getsuffixes    = suffixes,
  getbestsuffix  = bestsuffix,
  setcurrent     = setcurrent,
}
