-- Lua code used by macros.forge and derivatives

local fedora = require "fedora.common"

-- Computes the suffix of a version string, removing vprefix if it matches
-- For example with vprefix 1.2.3: 1.2.3.rc2 → .rc2 but 1.2.30 → 1.2.30 not 0
local function versionsuffix(vstring, vprefix)
  if (vstring:sub(1, #vprefix) == vprefix) and
     (not string.match(vstring:sub(#vprefix + 1), "^%.?%d")) then
    return vstring:sub(#vprefix + 1)
  else
    return vstring
  end
end

-- Check if an identified url is sane
local function check_forge_url(url, id, silent)
  local checkedurl  = nil
  local checkedid   = nil
  local urlpatterns = {
    gitlab = {
      pattern     = 'https://[^/]+/[^/]+/[^/#?]+',
      description = 'https://(…[-.])gitlab[-.]…/owner/repo'},
    pagure = {
      pattern     = 'https://[^/]+/[^/#?]+',
      description = 'https://pagure.io/repo'},
    pagure_ns = {
      pattern     = 'https://[^/]+/[^/]+/[^/#?]+',
      description = 'https://pagure.io/namespace/repo'},
    pagure_fork = {
      pattern     = 'https://[^/]+/fork/[^/]+/[^/#?]+',
      description = 'https://pagure.io/fork/owner/repo'},
    pagure_ns_fork = {
      pattern     = 'https://[^/]+/fork/[^/]+/[^/]+/[^/#?]+',
      description = 'https://pagure.io/fork/owner/namespace/repo'},
    ["gitea.com"] = {
      pattern     = 'https://[^/]+/[^/]+/[^/#?]+',
      description = 'https://gitea.com/owner/repo'},
    github = {
      pattern     = 'https://[^/]+/[^/]+/[^/#?]+',
      description = 'https://(…[-.])github[-.]…/owner/repo'},
    ["code.googlesource.com"] = {
      pattern     = 'https://code.googlesource.com/[^#?]*[^/#?]+',
      description = 'https://code.googlesource.com/…/repo'},
    ["bitbucket.org"] = {
      pattern     = 'https://[^/]+/[^/]+/[^/#?]+',
      description = 'https://bitbucket.org/owner/repo'}}
  if (urlpatterns[id] ~= nil) then
    checkedurl = url:match(urlpatterns[id]["pattern"])
    if (checkedurl == nil) then
      if not silent then
        fedora.error(id .. " URLs must match " ..
                     urlpatterns[id]["description"] .. "!")
      end
    else
      checkedid = id
    end
  end
  return checkedurl, checkedid
end

-- Check if an url matches a known forge
local function idforge(url, silent)
  local forge_url = nil
  local forge    = nil
  if (url ~= "") then
    forge = url:match("^[^:]+://([^/]+)/")
    if (forge == nil) then
      if not silent then
        fedora.error("URLs must include a protocol such as https:// " ..
                     "and a path starting with /!")
      end
    else
      if (forge == "pagure.io") then
        if     url:match("[^:]+://pagure.io/fork/[^/]+/[^/]+/[^/]+") then
          forge = "pagure_ns_fork"
        elseif url:match("[^:]+://pagure.io/fork/[^/]+/[^/]+") then
          forge = "pagure_fork"
        elseif url:match("[^:]+://pagure.io/[^/]+/[^/]+") then
          forge = "pagure_ns"
        elseif url:match("[^:]+://pagure.io/[^/]+") then
          forge = "pagure"
        end
      elseif forge:match("^gitlab[%.-]") or forge:match("[%.-]gitlab[%.]") then
        forge = "gitlab"
      elseif forge:match("^github[%.-]") or forge:match("[%.-]github[%.]") then
        forge = "github"
      end
      forge_url, forge = check_forge_url(url, forge, silent)
    end
  end
  return forge_url, forge
end

-- The rpm forge variables, split by initialisation stage:
local vars = {
  -- key variables that may control a declaration block
  key      = {"forge_url"},
  -- variables, used to compute a SCM reference
  ref      = {"tag", "commit", "branch", "version", "ref"},
  -- other computed or recomputed variables
  computed = {"tag", "forge_source", "forge_setup_args",
              "forge_patch_mode", "forge_patch_level",
              "fileref", "extractdir", "time", "date", "distprefix"},
}

-- Return all the suffixes for which one of the forge key variables is set
local function suffixes()
  return fedora.allsuffixes(vars["key"])
end

-- Executes the f function in a loop
--   – the loop is controled by index (a string):
--     — if the index value given to floop is nil, loop over the whole
--       range. Use readflag() to process rpm flag arguments safely.
--     — otherwise execute f for the specified index value only.
--   – f is passed index then otherargs as arguments.
--   – the index range is controlled by suffixes()
local function floop(f, index, otherargs, verbose)
  fedora.floop(f, index, suffixes(verbose), otherargs)
end

-- %forge_init core
local function init(suffix, verbose, informative, silent)
  local ismain = (suffix == "") or (suffix == "0")
  local rules = {
    default = {
      scm          = "git",
      archive_ext  = "tar.bz2",
      repo         = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "^[^:]+://[^/]+/[^/]+/([^/?#]+)"))}',
      archive_name = "%{repo"         .. suffix .. "}-%{ref"           .. suffix .. "}",
      topdir       = "%{archive_name"  .. suffix .. "}" },
    gitlab = {
      archive_url  = "%{forge_url"     .. suffix .. "}/-/archive/%{ref" .. suffix .. "}/%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}" },
    pagure = {
      archive_ext  = "tar.gz",
      repo         = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "^[^:]+://[^/]+/([^/?#]+)"))}',
      archive_url  = "%{forge_url"     .. suffix .. "}/archive/%{ref"   .. suffix .. "}/%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}" },
    pagure_ns = {
      archive_ext  = "tar.gz",
      namespace    = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "^[^:]+://[^/]+/([^/]+)/[^/?#]+"))}',
      repo         = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "^[^:]+://[^/]+/[^/]+/([^/?#]+)"))}',
      archive_name = "%{namespace"    .. suffix .. "}-%{repo"          .. suffix .. "}-%{ref"         .. suffix .. "}",
      archive_url  = "%{forge_url"     .. suffix .. "}/archive/%{ref"   .. suffix .. "}/%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}" },
    pagure_fork = {
      archive_ext  = "tar.gz",
      owner        = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "https://[^/]+/fork/([^/]+)/[^/?#]+"))}',
      repo         = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "https://[^/]+/fork/[^/]+/([^/?#]+)"))}',
      archive_name = "%{owner"        .. suffix .. "}-%{repo"          .. suffix .. "}-%{ref"         .. suffix .. "}",
      archive_url  = "%{forge_url"     .. suffix .. "}/archive/%{ref"   .. suffix .. "}/%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}" },
    pagure_ns_fork = {
      owner        = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "https://[^/]+/fork/([^/]+)/[^/]+/[^/?#]+"))}',
      namespace    = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "https://[^/]+/fork/[^/]+/([^/]+)/[^/?#]+")}',
      repo         = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "https://[^/]+/fork/[^/]+/[^/]+/([^/?#]+)")}',
      archive_name = "%{owner"        .. suffix .. "}-%{namespace"     .. suffix .. "}-%{repo"        .. suffix .. "}-%{ref"        .. suffix .. "}",
      archive_url  = "%{forge_url"     .. suffix .. "}/archive/%{ref"   .. suffix .. "}/%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}" },
    ["gitea.com"]  = {
      archive_ext  = "tar.gz",
      archive_name = "%{fileref"      .. suffix .. "}",
      archive_url  = "%{forge_url"     .. suffix .. "}/archive/%{ref"   .. suffix .. "}.%{archive_ext" .. suffix .. "}",
      topdir       = "%{repo}" },
    github = {
      archive_ext  = "tar.gz",
      archive_name = "%{repo"         .. suffix .. "}-%{fileref"       .. suffix .. "}",
      archive_url  = "%{forge_url"     .. suffix .. "}/archive/%{ref"   .. suffix .. "}/%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}" },
    ["code.googlesource.com"] = {
      archive_ext  = "tar.gz",
      repo         = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "^[^:]+://.+/([^/?#]+)"))}',
      archive_url  = "%{forge_url"     .. suffix .. "}/+archive/%{ref"  .. suffix .. "}.%{archive_ext"  .. suffix .. "}",
      topdir       = "" },
    ["bitbucket.org"] = {
      short_commit = '%{lua:print(string.sub(rpm.expand("%{commit'     .. suffix .. '}"), 1, 12))}',
      owner        = '%{lua:print(string.match(rpm.expand("%{forge_url' .. suffix .. '}"), "^[^:]+://[^/]+/([^/?#]+)"))}',
      archive_name = "%{owner"        .. suffix .. "}-%{repo"          .. suffix .. "}-%{short_commit" .. suffix .. "}",
      archive_url  = "%{forge_url"     .. suffix .. "}/get/%{ref"       .. suffix .. "}.%{archive_ext"  .. suffix .. "}" } }
  vars["rules"] = {}
  for _, v in pairs(rules) do
    local l = {}
    for k, _ in pairs(v) do
      table.insert(l, k)
    end
    vars["rules"] = fedora.mergelists({vars["rules"], l})
  end
  local vars_all = {}
  for k, _ in pairs(vars) do
    vars_all   = fedora.mergelists({vars_all, vars[k]})
  end
  vars["all"] = vars_all
  if ismain then
    fedora.zalias(vars["all"])
  end
  local spec = {}
  for _, v in ipairs(vars["all"]) do
    spec[v] = fedora.read(v .. suffix)
  end
  -- Compute the reference of the object to fetch
  local isrelease = false
  local       ref
  -- Since tag is well defined in git and version – not, tag declaration
  -- takes precedence over version. Commit *is* well defined but
  -- human-unfriendly. Packaging a moving branch is quite a bad idea, but since
  -- forges export branch references, better support those than have packagers
  -- hack up workarounds
  if     spec["tag"]    then ref = "%{?tag"     .. suffix .. "}"
  elseif spec["commit"] then ref = "%{?commit"  .. suffix .. "}"
  elseif spec["branch"] then ref = "%{?branch"  .. suffix .. "}"
  else                       ref = "%{?version" .. suffix .. "}"
                       isrelease = true
  end
  -- Do not set ref yet because it needs some v munging
  if not spec["ref"] and (rpm.expand(ref) == "") then
    fedora.error([[
You need to set one of:
  – %%{tag]]     .. suffix .. [[}
  – %%{version]] .. suffix .. [[}
  – %%{commit]]  .. suffix .. [[}
with your:
%%{forge_url]] .. suffix .. [[} %{?forge_url]] .. suffix .. [[}
declaration.
]])
  end
  local  forge_url = spec["forge_url"]
  local  forge
  forge_url, forge = idforge(forge_url, silent)
  if forge then
    fedora.set("forge_url" .. suffix, forge_url, verbose)
    -- Custom processing of quirky forges
    if (forge == "github") then
      -- Workaround the way GitHub injects "v"s before some version strings
      -- (but not all!). To package one of the minority of sane GitHub projects
      -- that do not munge their version strings set tag to %{version} in your
      -- spec
      local fileref = ref
      if (ref == "%{?version"  .. suffix .. "}") then
        ref = "v" .. ref
      elseif (fileref ~= "%{?commit" .. suffix .. "}") and
             string.match(rpm.expand(fileref), "^v[%d]") then
        fileref = string.gsub(rpm.expand(fileref), "^v", "")
      elseif (string.match(rpm.expand(fileref), "/")) then
        fileref = string.gsub(rpm.expand(fileref), "/", "-")
      end
      fedora.safeset("fileref" .. suffix, fileref, verbose)
    elseif (forge == "gitea.com") then
      -- Workaround the way gitea mangles /s in ref names
      local fileref = ref
      fileref = string.gsub(rpm.expand(fileref), "/", "-")
      fedora.safeset("fileref" .. suffix, fileref, verbose)
    elseif (forge == "code.googlesource.com") then
      if (ref == "%{?version"  .. suffix .. "}") then
        ref = "v" .. ref
      end
    elseif (forge == "bitbucket.org") then
      if spec["commit"] then
        fedora.error("All BitBucket URLs require commit value knowledge: " ..
                     "you need to set %{commit}!")
      end
    end
    fedora.safeset("ref" .. suffix, ref, verbose)
    -- Mass setting of the remaining variables
    for k,v in pairs(rules[forge]) do
      fedora.safeset(k .. suffix, rules[forge][k], verbose)
    end
    for k,v in pairs(rules["default"]) do
      if (rules[forge][k] == nil) then
        fedora.safeset(k .. suffix, rules["default"][k], verbose)
      end
    end
    -- Update lua copy with the result
    for _, v in ipairs(vars["rules"]) do
      spec[v] = fedora.read(v .. suffix)
    end
  end
  fedora.safeset("source_url", "%{forge_url" .. suffix .. "}", verbose)
  -- Source URL processing (computing the forge_source spec variable)
  local forge_source = "%{archive_url"  .. suffix .. "}"
  local archive_file = "%{archive_name" .. suffix .. "}.%{archive_ext" .. suffix .. "}"
  if (string.match(rpm.expand(forge_source), "/([^/]+)$")
                ~= rpm.expand(archive_file)) then
    forge_source = forge_source .. "#/" .. archive_file
  end
  fedora.safeset("forge_source" .. suffix, forge_source, verbose)
  -- Setup processing      (computing the forgesetup and extractdir variables)
  local forge_setup_args = "-n %{extractdir" .. suffix .. "}"
  local extractdir     = "%{topdir"        .. suffix .. "}"
  if (spec["topdir"] == "") then
    forge_setup_args     = "-c " .. forge_setup_args
    extractdir         = "%{archive_name"   .. suffix .. "}"
  end
  if not ismain then
    if (spec["topdir"] == "") then
      forge_setup_args = "-T -D -a " .. suffix .. " " .. forge_setup_args
    else
      forge_setup_args = "-T -D -b " .. suffix .. " " .. forge_setup_args
    end
  end
  fedora.safeset("forge_setup_args" .. suffix, forge_setup_args, verbose)
  fedora.safeset("extractdir"     .. suffix, extractdir, verbose)
  -- %apply_patch setup; patch and quilt are scms? Really?
  -- Due to the design of %apply_patch we will need to set %{__scm} from
  -- %{forge_patch_modeX} later. Therefore, expand the fallback to avoid loops
  -- and inter-block side effects
  fedora.safeset("forge_patch_mode" .. suffix, fedora.read("__scm"), verbose)
  -- That’s the default in rpm guides since forever, but people *still* insist
  -- in making it a parameter instead of defining a %_default_patch_level like
  -- has been done for fuzz
  fedora.safeset("forge_patch_level" .. suffix, "-p 1", verbose)
  -- Date processing. For backwards compatibility reasons we have a %{dateX}
  -- variable, that used to be sufficient before people started doing
  -- processing that requires %{timeX} and second resolution (we could do
  -- nanosecond resolution but that would be a joke given the state of
  -- timestamp handling in existing SCMs).
  if spec["date"] then
    local t = '%([[ -n "%{date' .. suffix .. '}" ]] && ' ..
              'date -u --iso-8601=seconds -d %{date' .. suffix .. '})'
    fedora.safeset(  "time" .. suffix, t)
  end
  local t = "%([ -r %{_sourcedir}/" .. archive_file .. " ] && " ..
            "date -u --iso-8601=seconds -r %{_sourcedir}/" .. archive_file ..
            ")"
  fedora.safeset(  "time" .. suffix, t)
  spec["time"] = fedora.read("time" .. suffix)
  if spec["time"] then
    local d = '%([[ -n "%{time' .. suffix .. '}" ]] && ' ..
              'date -u +%Y%m%d -d %{time' .. suffix .. '})'
    fedora.safeset(  "date" .. suffix, d)
  end
  -- dist processing       (computing the correct prefix for snapshots)
  local distprefix = ""
  if not isrelease then
    distprefix = string.lower(rpm.expand(ref))
    distprefix = distprefix:gsub("/", ".")
    if     (ref == "%{?commit" .. suffix .. "}") then
      distprefix = distprefix:sub(1, 7)
    elseif (ref ~= "%{?branch" .. suffix .. "}") then
      distprefix = distprefix:gsub("[%p%s]+", ".")
      distprefix = distprefix:gsub("^" ..
                   string.lower(rpm.expand("%{?repo}")) .. "%.?", "")
      local    v = string.gsub( rpm.expand("%{version}"), "[%p%s]+", ".")
      for _, p in ipairs({ '', 'v', 'v.', 'version', 'version.',
                          'tags.v', 'tags.v.' }) do
        distprefix = versionsuffix(distprefix, p .. v)
      end
      distprefix = distprefix:gsub("^%.", "")
    end
    if (distprefix ~= "") then
      distprefix = "." .. "%{scm"  .. suffix .. "}" ..
                   "%{?date" .. suffix .. "}" .. distprefix
    end
  end
  if  spec["version"] and
     (spec["version"] ~= "") and
     (spec["version"] ~= "0") and
     (spec["version"] ~= "0.0.0") and
     (spec["version"] ~= fedora.read("version")) then
    distprefix = ".%{version" .. suffix .. "}" .. distprefix
  end
  if (rpm.expand(distprefix) ~= "") then
    if not ismain then
      distprefix = distprefix:gsub("^%.", ".s")
    end
    fedora.safeset ("distprefix"    .. suffix, distprefix, verbose)
  end
  if ismain then
    fedora.zalias(fedora.mergelists(vars["key"], vars["rules"],
                                    vars["computed"]), verbose)
  end
  -- Final spec variable summary if the macro was called with -i
  if informative then
    fedora.echo("Packaging variables read or set by %%forge_init")
    fedora.echovars(vars["all"], suffix)
    fedora.echovars({"dist"},"")
    fedora.echo([[
  Unless explicitly set, snapshot timestamp will be computed once
  %%{_sourcedir}/]] .. archive_file .. [[
  is available.
]])
  end
end

-- %forge_sources core
local function source(suffix)
  print(rpm.expand("%{?forge_source"    .. suffix .. "}\n"))
end

-- %forge_patches core
local function patchlist(suffix)
  print(rpm.expand("%{?forge_patchlist" .. suffix .. "}\n"))
end

-- This is constrained by the design of %apply_patch
local function apply_patches(suffix, verbose)
  local quiet = ""
  if not verbose then
    quiet = "-q "
  end
  for _, patch in ipairs(fedora.readlines("forge_patchlist" .. suffix)) do
    local ps = fedora.patchsuffix(patch)
    if ps then
      for _, o in ipairs({"patch_mode", "patch_level"}) do
        fedora.safeset(o .. ps, "%{forge_" .. o .. suffix .. "}")
      end
      -- %apply_patch does not have the notion of a disposable local currentfoo-like
      -- control variable, you need to override the main one
      fedora.set("__scm", "%{patch_mode" .. ps .. "}")
      -- %apply_patch requires specifying the patch to apply in 3 different forms!
      -- As a filename, as a full path, and as a patch suffix
      print(rpm.expand(
        "%apply_patch " ..  quiet .. "%{patch_level" .. ps .. "} " ..
        "-m " .. patch .. " %{_sourcedir}/" .. patch .. " " .. ps .. "\n"))
    end
  end
end

local function env(suffix, verbose)
  fedora.set_current(vars["all"], suffix, verbose)
  fedora.set_verbose(verbose)
end

-- %forge_prep core
local function prep(suffix, verbose)
  env(suffix, verbose)
  print(rpm.expand([[
%setup %{?__current_quiet} %{__current_forge_setup_args}
]]))
  apply_patches(suffix, verbose)
end

return {
  suffixes  = suffixes,
  floop     = floop,
  init      = init,
  env       = env,
  source    = source,
  patchlist = patchlist,
  prep      = prep,
  -- deprecated aliases
  meta      = init,
  setup     = prep,
}
